@@ -13,7 +13,7 @@ BEGIN {
 
 use Graph::AdjacencyMap qw(:flags :fields);
 
-our $VERSION = '0.9714';
+our $VERSION = '0.9715';
 
 require 5.006; # Weak references are absolutely required.
 
@@ -68,7 +68,6 @@ sub Infinity () { $Inf }
 # The edges are always in a "map".
 # The defined flags for maps are:
 # - _COUNT for countedness: more than one instance
-# - _HYPER for hyperness: a different number of "coordinates" than usual;
 #   expects one for vertices and two for edges
 # - _UNORD for unordered coordinates (a set): if _UNORD is not set
 #   the coordinates are assumed to be meaningfully ordered
@@ -170,9 +169,9 @@ sub new {
     _opt(\%opt, \$eflags,
 	 countedged	=> _COUNT,
 	 multiedged	=> _MULTI,
-	 hyperedged	=> _HYPER,
 	 undirected	=> _UNORD,
 	);
+    my $is_hyper = delete $opt{hyperedged};
 
     _opt(\%opt, \$gflags,
 	 unionfind     => _UNIONFIND,
@@ -207,8 +206,8 @@ sub new {
     $g->[ _V ] = $vflags ?
 	_am_heavy($vflags, 1) :
 	    _am_light($vflags, 1);
-    $g->[ _E ] = ($eflags & ~_UNORD) ?
-	_am_heavy($eflags, 2) :
+    $g->[ _E ] = ($is_hyper or $eflags & ~_UNORD) ?
+	_am_heavy($eflags, $is_hyper ? undef : 2) :
 	    _am_light($eflags, 2);
 
     $g->add_vertices(@V) if @V;
@@ -242,7 +241,7 @@ sub __stringified { $_[0]->[ _V ]->_is_STR   }
 
 sub countedged    { $_[0]->[ _E ]->_is_COUNT }
 sub multiedged    { $_[0]->[ _E ]->_is_MULTI }
-sub hyperedged    { $_[0]->[ _E ]->_is_HYPER }
+sub hyperedged    { !defined $_[0]->[ _E ]->[ _arity ] }
 sub undirected    { $_[0]->[ _E ]->_is_UNORD }
 
 sub directed { ! $_[0]->[ _E ]->_is_UNORD }
@@ -272,8 +271,8 @@ sub add_vertex {
 	push @_, _GEN_ID;
 	goto &add_vertex_by_id;
     }
-    return $g if !&is_countvertexed and $V->has_path( $_[1] );
-    $V->set_path( @_[1..$#_] );
+    return $g if !&is_countvertexed and $V->has_path([$_[1]]);
+    $V->set_path([@_[1..$#_]]);
     $g->[ _G ]++;
     &_union_find_add_vertex if &has_union_find;
     return $g;
@@ -283,7 +282,7 @@ sub has_vertex {
     my $g = $_[0];
     my $V = $g->[ _V ];
     return exists $V->[ _s ]->{ $_[1] } if ($V->[ _f ] & _LIGHT);
-    $V->has_path( $_[1] );
+    $V->has_path([$_[1]]);
 }
 
 sub _vertices05 {
@@ -334,10 +333,11 @@ sub add_edge {
 	push @_, _GEN_ID;
 	goto &add_edge_by_id;
     }
-    my @e = &_vertex_ids_ensure;
-    $g->[ _E ]->set_path( @e );
+    my @i = &_vertex_ids_ensure;
+    @i = sort @i if &is_undirected;
+    $g->[ _E ]->set_path( \@i );
     $g->[ _G ]++;
-    $g->_union_find_add_edge( @e ) if &has_union_find;
+    $g->_union_find_add_edge( @i ) if &has_union_find;
     return $g;
 }
 
@@ -374,11 +374,11 @@ sub _vertex_ids_maybe_ensure {
 sub has_edge {
     my $g = $_[0];
     my $E = $g->[ _E ];
-    my $Ef = $E->[ _f ];
-    return 0 if !($Ef & _HYPER) and @_ != 3;
+    my ($Ef, $Ea) = @$E[ _f, _arity ];
+    return 0 if defined($Ea) and @_ != $Ea + 1;
     return 0 if (my @i = &_vertex_ids) != @_ - 1;
-    return $E->get_ids_by_paths([ \@i ]) if !(@i == 2 && !($Ef & (_HYPER|_REF|_UNIQ))); # Slow path.
-    @i = sort @i if ($Ef & _UNORD);
+    @i = sort @i if &is_undirected;
+    return $E->get_ids_by_paths([ \@i ]) if !(@i == 2 && !(!defined($Ea) or $Ef & (_REF|_UNIQ))); # Slow path.
     my $s = $E->[ _s ];
     return exists $s->{ $i[0] } &&
 	   exists $s->{ $i[0] }->{ $i[1] } ? 1 : 0;
@@ -389,6 +389,7 @@ sub any_edge {
     my $E = $g->[ _E ];
     my $Ef = $E->[ _f ];
     return 0 if (my @i = &_vertex_ids) != @_ - 1;
+    @i = sort @i if &is_undirected;
     &_edge_cache;
     my $N0 = $g->[ _S ][0];
     my $s = $N0->{ $i[0] };
@@ -415,8 +416,7 @@ sub edges {
 }
 
 sub has_edges {
-    my $g = shift;
-    scalar $g->[ _E ]->has_paths( @_ );
+    scalar $_[0]->[ _E ]->has_paths;
 }
 
 ###
@@ -425,10 +425,10 @@ sub has_edges {
 
 sub add_vertex_by_id {
     &expect_multivertexed;
-    my $g = $_[0];
+    my ($g, $v, $id) = @_;
     my $V = $g->[ _V ];
-    return $g if $V->has_path_by_multi_id( @_[1..$#_] );
-    $V->set_path_by_multi_id( @_[1..$#_] );
+    return $g if $V->has_path_by_multi_id( my @args = ([$v], $id) );
+    $V->set_path_by_multi_id( @args );
     $g->[ _G ]++;
     &_union_find_add_vertex if &has_union_find;
     return $g;
@@ -436,8 +436,8 @@ sub add_vertex_by_id {
 
 sub add_vertex_get_id {
     &expect_multivertexed;
-    my $g = $_[0];
-    my $id = $g->[ _V ]->set_path_by_multi_id( $_[1], _GEN_ID );
+    my ($g, $v) = @_;
+    my $id = $g->[ _V ]->set_path_by_multi_id( [$v], _GEN_ID );
     $g->[ _G ]++;
     &_union_find_add_vertex if &has_union_find;
     return $id;
@@ -445,18 +445,18 @@ sub add_vertex_get_id {
 
 sub has_vertex_by_id {
     &expect_multivertexed;
-    $_[0]->[ _V ]->has_path_by_multi_id( @_[1..$#_] );
+    my ($g, $v, $id) = @_;
+    $g->[ _V ]->has_path_by_multi_id( [$v], $id );
 }
 
 sub delete_vertex_by_id {
     &expect_multivertexed;
     &expect_non_unionfind;
-    my $g = shift;
-    my $V = $g->[ _V ];
-    return unless $V->has_path_by_multi_id( @_ );
+    my ($g, $v, $id) = @_;
+    return $g unless &has_vertex_by_id;
     # TODO: what to about the edges at this vertex?
     # If the multiness of this vertex goes to zero, delete the edges?
-    $V->del_path_by_multi_id( @_ );
+    $g->[ _V ]->del_path_by_multi_id( [$v], $id );
     $g->[ _G ]++;
     return $g;
 }
@@ -464,16 +464,18 @@ sub delete_vertex_by_id {
 sub get_multivertex_ids {
     &expect_multivertexed;
     my $g = shift;
-    $g->[ _V ]->get_multi_ids( @_ );
+    $g->[ _V ]->get_multi_ids( \@_ );
 }
 
 sub add_edge_by_id {
     &expect_multiedged;
     my $g = $_[0];
     my @i = &_vertex_ids_ensure_multi;
-    $g->[ _E ]->set_path_by_multi_id( @i );
+    my $id = pop @i;
+    @i = sort @i if &is_undirected;
+    $g->[ _E ]->set_path_by_multi_id( \@i, $id );
     $g->[ _G ]++;
-    $g->_union_find_add_edge( @i[0..$#i-1] ) if &has_union_find;
+    $g->_union_find_add_edge( @i ) if &has_union_find;
     return $g;
 }
 
@@ -481,7 +483,8 @@ sub add_edge_get_id {
     &expect_multiedged;
     my $g = $_[0];
     my @i = &_vertex_ids_ensure;
-    my $id = $g->[ _E ]->set_path_by_multi_id( @i, _GEN_ID );
+    @i = sort @i if &is_undirected;
+    my $id = $g->[ _E ]->set_path_by_multi_id( \@i, _GEN_ID );
     $g->[ _G ]++;
     $g->_union_find_add_edge( @i ) if &has_union_find;
     return $id;
@@ -491,8 +494,10 @@ sub has_edge_by_id {
     &expect_multiedged;
     my $g = $_[0];
     my @i = &_vertex_ids_multi;
-    return 0 if @i == 0 && @_ - 1;
-    $g->[ _E ]->has_path_by_multi_id( @i );
+    return 0 if @i < @_ - 2;
+    my $id = pop @i;
+    @i = sort @i if &is_undirected;
+    $g->[ _E ]->has_path_by_multi_id( \@i, $id );
 }
 
 sub delete_edge_by_id {
@@ -500,17 +505,20 @@ sub delete_edge_by_id {
     &expect_non_unionfind;
     my $g = $_[0];
     my $E = $g->[ _E ];
-    return unless my @i = &_vertex_ids_multi;
-    return unless $E->has_path_by_multi_id( @i );
-    $E->del_path_by_multi_id( @i );
+    my @i = &_vertex_ids_multi;
+    return if @i < @_ - 2;
+    my $id = pop @i;
+    @i = sort @i if &is_undirected;
+    return unless $E->has_path_by_multi_id( my @args = (\@i, $id) );
+    $E->del_path_by_multi_id( @args );
     $g->[ _G ]++;
     return $g;
 }
 
 sub get_multiedge_ids {
     &expect_multiedged;
-    return unless my @i = &_vertex_ids;
-    $_[0]->[ _E ]->get_multi_ids( @i );
+    return unless @_-1 == (my @i = &_vertex_ids);
+    $_[0]->[ _E ]->get_multi_ids( \@i );
 }
 
 ###
@@ -664,7 +672,8 @@ sub delete_edge {
     &expect_non_unionfind;
     my $g = $_[0];
     my @i = &_vertex_ids;
-    return $g unless @i and $g->[ _E ]->del_path( @i );
+    @i = sort @i if &is_undirected;
+    return $g unless @i and $g->[ _E ]->del_path( \@i );
     $g->[ _G ]++;
     return $g;
 }
@@ -674,25 +683,26 @@ sub delete_vertex {
     my $g = $_[0];
     return $g if @_ != 2;
     my $V = $g->[ _V ];
-    return $g unless $V->has_path( $_[1] );
+    return $g unless $V->has_path([$_[1]]);
     # TODO: _edges_at is excruciatingly slow (rt.cpan.org 92427)
     my $E = $g->[ _E ];
-    $E->del_path( @$_ ) for &_edges_at;
-    $V->del_path( $_[1] );
+    $E->del_path( $_ ) for &_edges_at;
+    $V->del_path([$_[1]]);
     $g->[ _G ]++;
     return $g;
 }
 
 sub get_vertex_count {
     my $g = shift;
-    $g->[ _V ]->_get_path_count( @_ ) || 0;
+    $g->[ _V ]->_get_path_count( \@_ ) || 0;
 }
 
 sub get_edge_count {
     my $g = $_[0];
     my @i = &_vertex_ids;
     return 0 unless @i;
-    $g->[ _E ]->_get_path_count( @i ) || 0;
+    @i = sort @i if &is_undirected;
+    $g->[ _E ]->_get_path_count( \@i ) || 0;
 }
 
 sub delete_vertices {
@@ -941,6 +951,14 @@ for my $entity (qw(vertex edge)) {
     }
 }
 
+sub _munge_args {
+    my ($is_multi, $is_undirected, @args) = @_;
+    return \@args if !$is_undirected and !$is_multi;
+    return [ sort @args ] if !$is_multi;
+    my $id = pop @args;
+    ($is_undirected ? [ sort @args ] : \@args, $id);
+}
+
 sub _set_attribute {
     my ($is_multi, $entity, $offset, $args) = splice @_, -4, 4;
     my $value = pop;
@@ -948,6 +966,7 @@ sub _set_attribute {
     no strict 'refs';
     &{ 'add_' . $entity . ($is_multi ? '_by_id' : '') } unless &{ 'has_' . $entity . ($is_multi ? '_by_id' : '') };
     my @args = ($entity eq 'edge') ? &$args : @_[1..$#_];
+    @args = _munge_args($is_multi, &is_undirected, @args);
     $_[0]->[ $offset ]->_set_path_attr( @args, $attr, $value );
 }
 
@@ -957,6 +976,7 @@ sub _set_attributes {
     no strict 'refs';
     &{ 'add_' . $entity . ($is_multi ? '_by_id' : '') } unless &{ 'has_' . $entity . ($is_multi ? '_by_id' : '') };
     my @args = ($entity eq 'edge') ? &$args : @_[1..$#_];
+    @args = _munge_args($is_multi, &is_undirected, @args);
     $_[0]->[ $offset ]->_set_path_attrs( @args, $attr );
 }
 
@@ -965,6 +985,7 @@ sub _has_attributes {
     no strict 'refs';
     return 0 unless &{ 'has_' . $entity . ($is_multi ? '_by_id' : '') };
     my @args = ($entity eq 'edge') ? &$args : @_[1..$#_];
+    @args = _munge_args($is_multi, &is_undirected, @args);
     $_[0]->[ $offset ]->_has_path_attrs( @args );
 }
 
@@ -974,6 +995,7 @@ sub _has_attribute {
     no strict 'refs';
     return 0 unless &{ 'has_' . $entity . ($is_multi ? '_by_id' : '') };
     my @args = ($entity eq 'edge') ? &$args : @_[1..$#_];
+    @args = _munge_args($is_multi, &is_undirected, @args);
     $_[0]->[ $offset ]->_has_path_attr( @args, $attr );
 }
 
@@ -982,6 +1004,7 @@ sub _get_attributes {
     no strict 'refs';
     return undef unless &{ 'has_' . $entity . ($is_multi ? '_by_id' : '') };
     my @args = ($entity eq 'edge') ? &$args : @_[1..$#_];
+    @args = _munge_args($is_multi, &is_undirected, @args);
     scalar $_[0]->[ $offset ]->_get_path_attrs( @args );
 }
 
@@ -991,6 +1014,7 @@ sub _get_attribute {
     my $attr = pop;
     return undef unless &{ 'has_' . $entity . ($is_multi ? '_by_id' : '') };
     my @args = ($entity eq 'edge') ? &$args : @_[1..$#_];
+    @args = _munge_args($is_multi, &is_undirected, @args);
     scalar $_[0]->[ $offset ]->_get_path_attr( @args, $attr );
 }
 
@@ -999,6 +1023,7 @@ sub _get_attribute_names {
     no strict 'refs';
     return unless &{ 'has_' . $entity . ($is_multi ? '_by_id' : '') };
     my @args = ($entity eq 'edge') ? &$args : @_[1..$#_];
+    @args = _munge_args($is_multi, &is_undirected, @args);
     $_[0]->[ $offset ]->_get_path_attr_names( @args );
 }
 
@@ -1007,6 +1032,7 @@ sub _get_attribute_values {
     no strict 'refs';
     return unless &{ 'has_' . $entity . ($is_multi ? '_by_id' : '') };
     my @args = ($entity eq 'edge') ? &$args : @_[1..$#_];
+    @args = _munge_args($is_multi, &is_undirected, @args);
     $_[0]->[ $offset ]->_get_path_attr_values( @args );
 }
 
@@ -1015,6 +1041,7 @@ sub _delete_attributes {
     no strict 'refs';
     return undef unless &{ 'has_' . $entity . ($is_multi ? '_by_id' : '') };
     my @args = ($entity eq 'edge') ? &$args : @_[1..$#_];
+    @args = _munge_args($is_multi, &is_undirected, @args);
     $_[0]->[ $offset ]->_del_path_attrs( @args );
 }
 
@@ -1024,6 +1051,7 @@ sub _delete_attribute {
     no strict 'refs';
     return undef unless &{ 'has_' . $entity . ($is_multi ? '_by_id' : '') };
     my @args = ($entity eq 'edge') ? &$args : @_[1..$#_];
+    @args = _munge_args($is_multi, &is_undirected, @args);
     $_[0]->[ $offset ]->_del_path_attr( @args, $attr );
 }
 
@@ -1096,8 +1124,11 @@ sub as_hashes {
             }
         }
     } else {
-        $e{ $_->[0] }{ $_->[1] } = $g->get_edge_attributes(@$_) || {}
-            for $g->edges;
+	for ($g->edges) {
+	    $e{ $_->[0] }{ $_->[1] } = $g->get_edge_attributes(@$_) || {};
+	    $e{ $_->[1] }{ $_->[0] } = $g->get_edge_attributes(@$_) || {}
+		if !$is_directed;
+	}
     }
     ( \%n, $is_hyper ? \@e : \%e );
 }
@@ -2712,30 +2743,19 @@ sub could_be_isomorphic {
 ###
 # Analysis functions.
 
-sub subgraph_by_radius
-{
-    my ($g, $n, $rad) = @_;
-
-    return unless defined $n && defined $rad && $rad >= 0;
-
-    my $r = (ref $g)->new;
-
-    return $r->add_vertex($n) if $rad == 0;
-
-    my %h;
-    $h{1} = [ [ $n, $g->successors($n) ] ];
-    for my $i (1..$rad) {
-	$h{$i+1} = [];
-	for my $arr (@{ $h{$i} }) {
-	    my ($p, @succ) = @{ $arr };
-	    for my $s (@succ) {
-		$r->add_edge($p, $s);
-		push(@{ $h{$i+1} }, [$s, $g->successors($s)]) if $i < $rad;
-	    }
-	}
+sub subgraph_by_radius {
+    my ($g, @v) = @_;
+    my $rad = pop @v;
+    return unless defined $rad && $rad >= 0 && @v;
+    my %got;
+    my @next = @got{ @v } = @v;
+    while ($rad--) {
+	my @succ = $g->successors(@next);
+	my %notseen; @notseen{ @succ } = @succ;
+	delete @notseen{ keys %got };
+	@next = @got{ keys %notseen } = values %notseen;
     }
-
-    return $r;
+    $g->subgraph([ values %got ]);
 }
 
 sub clustering_coefficient {
